/*
 *  VectorObject_jni.cpp
 *  WhirlyGlobeLib
 *
 *  Created by Steve Gifford on 6/2/14.
 *  Copyright 2011-2016 mousebird consulting
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

#import <jni.h>
#import "Maply_jni.h"
#import "com_mousebird_maply_VectorObject.h"
#import "WhirlyGlobe.h"
#import "Maply_utils_jni.h"

using namespace WhirlyKit;

JNIEXPORT void JNICALL Java_com_mousebird_maply_VectorObject_nativeInit
  (JNIEnv *env, jclass cls)
{
	VectorObjectClassInfo::getClassInfo(env,cls);
}

JNIEXPORT jobject JNICALL MakeVectorObject(JNIEnv *env,VectorObject *vec)
{
	VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo(env,"com/mousebird/maply/VectorObject");
	jobject newObj = classInfo->makeWrapperObject(env,NULL);
	VectorObject *inst = classInfo->getObject(env,newObj);
	if (inst)
		delete inst;
	classInfo->setHandle(env,newObj,vec);

	return newObj;
}

void Java_com_mousebird_maply_VectorObject_initialise
  (JNIEnv *env, jobject obj)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		VectorObject *inst = new VectorObject();
		classInfo->setHandle(env,obj,inst);
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::initialise()");
	}
}

static std::mutex disposeMutex;

void Java_com_mousebird_maply_VectorObject_dispose
  (JNIEnv *env, jobject obj)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        {
            std::lock_guard<std::mutex> lock(disposeMutex);
            VectorObject *inst = classInfo->getObject(env,obj);
            if (!inst)
                return;
            delete inst;

            classInfo->clearHandle(env,obj);
        }
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::dispose()");
	}
}

JNIEXPORT void JNICALL Java_com_mousebird_maply_VectorObject_addPoint
  (JNIEnv *env, jobject obj, jobject ptObj)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		VectorObject *vecObj = classInfo->getObject(env,obj);
		Point2d *pt = Point2dClassInfo::getClassInfo()->getObject(env,ptObj);
		if (!vecObj)
			return;

		VectorPointsRef pts = VectorPoints::createPoints();
		pts->pts.push_back(GeoCoord(pt->x(),pt->y()));
		pts->initGeoMbr();
		vecObj->shapes.insert(pts);
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::addPoint()");
	}
}

JNIEXPORT void JNICALL Java_com_mousebird_maply_VectorObject_addLinear
  (JNIEnv *env, jobject obj, jobjectArray ptsObj)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		VectorObject *vecObj = classInfo->getObject(env,obj);
		if (!vecObj)
			return;

		VectorLinearRef lin = VectorLinear::createLinear();

		int count = env->GetArrayLength(ptsObj);
		if (count == 0)
			return;
		for (int ii=0;ii<count;ii++)
		{
			jobject ptObj = env->GetObjectArrayElement(ptsObj,ii);
			Point2d *pt = Point2dClassInfo::getClassInfo()->getObject(env,ptObj);
			lin->pts.push_back(GeoCoord(pt->x(),pt->y()));
            env->DeleteLocalRef(ptObj);
		}
		lin->initGeoMbr();
		vecObj->shapes.insert(lin);
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::addLinear()");
	}
}

JNIEXPORT void JNICALL Java_com_mousebird_maply_VectorObject_addAreal___3Lcom_mousebird_maply_Point2d_2
  (JNIEnv *env, jobject obj, jobjectArray ptsObj)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		VectorObject *vecObj = classInfo->getObject(env,obj);
		if (!vecObj)
			return;

		VectorArealRef ar = VectorAreal::createAreal();
		ar->loops.resize(1);

		int count = env->GetArrayLength(ptsObj);
		if (count == 0)
			return;
		for (int ii=0;ii<count;ii++)
		{
			jobject ptObj = env->GetObjectArrayElement(ptsObj,ii);
			Point2d *pt = Point2dClassInfo::getClassInfo()->getObject(env,ptObj);
			ar->loops[0].push_back(GeoCoord(pt->x(),pt->y()));
            env->DeleteLocalRef(ptObj);
		}
		ar->initGeoMbr();
		vecObj->shapes.insert(ar);
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::addAreal()");
	}
}

JNIEXPORT void JNICALL Java_com_mousebird_maply_VectorObject_addAreal___3Lcom_mousebird_maply_Point2d_2_3_3Lcom_mousebird_maply_Point2d_2
(JNIEnv *env, jobject obj, jobjectArray outerLoopObj, jobjectArray holesArray)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        if (!vecObj)
            return;
        Point2dClassInfo *point2dClassInfo = Point2dClassInfo::getClassInfo();
        
        VectorArealRef ar = VectorAreal::createAreal();
        ar->loops.resize(1+env->GetArrayLength(holesArray));

        for (int loop=0;loop<ar->loops.size();loop++)
        {
            jobjectArray ptsObj = loop == 0 ? outerLoopObj : (jobjectArray)env->GetObjectArrayElement(holesArray,loop-1);
            
            int count = env->GetArrayLength(ptsObj);
            if (count == 0)
                continue;
            for (int ii=0;ii<count;ii++)
            {
                jobject ptObj = env->GetObjectArrayElement(ptsObj,ii);
                Point2d *pt = point2dClassInfo->getObject(env,ptObj);
                ar->loops[loop].push_back(GeoCoord(pt->x(),pt->y()));
                env->DeleteLocalRef(ptObj);
            }
            
            if (loop > 0)
                env->DeleteLocalRef(ptsObj);
        }

        ar->initGeoMbr();
        vecObj->shapes.insert(ar);
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::addArealWithHoles()");
    }
}

jboolean Java_com_mousebird_maply_VectorObject_fromGeoJSON
  (JNIEnv *env, jobject obj, jstring jstr)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		VectorObject *vecObj = classInfo->getObject(env,obj);
		if (!vecObj)
			return false;

		const char *cStr = env->GetStringUTFChars(jstr,0);
		if (!cStr)
			return false;
		std::string jsonStr(cStr);
		env->ReleaseStringUTFChars(jstr, cStr);

		bool ret = vecObj->fromGeoJSON(jsonStr);

		return ret;
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::fromGeoJSON()");
	}
    
    return false;
}

JNIEXPORT jboolean JNICALL Java_com_mousebird_maply_VectorObject_fromShapeFile
(JNIEnv *env, jobject obj, jstring jstr)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        if (!vecObj)
            return false;
        
        const char *cStr = env->GetStringUTFChars(jstr,0);
        if (!cStr)
            return false;
        std::string jsonStr(cStr);
        env->ReleaseStringUTFChars(jstr, cStr);
        
        bool ret = vecObj->fromShapeFile(jsonStr);
        
        return true;
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::fromShapeFile()");
    }
    
    return false;
}

JNIEXPORT jobject JNICALL Java_com_mousebird_maply_VectorObject_FromGeoJSONAssembly
  (JNIEnv *env, jclass vecObjClass, jstring jstr)
{
	try
	{
		const char *cStr = env->GetStringUTFChars(jstr,0);
		if (!cStr)
			return NULL;
		std::string jsonStr(cStr);
		env->ReleaseStringUTFChars(jstr, cStr);

		std::map<std::string,VectorObject *> vecData;
		bool ret = VectorObject::FromGeoJSONAssembly(jsonStr,vecData);

		if (ret)
		{
			JavaHashMapInfo *hashMapClassInfo = JavaHashMapInfo::getClassInfo(env);
			jobject hashMap = hashMapClassInfo->makeHashMap(env);
			for (std::map<std::string,VectorObject *>::iterator it = vecData.begin();
					it != vecData.end(); ++it)
			{
				jstring key = env->NewStringUTF(it->first.c_str());
				jobject vecObj = MakeVectorObject(env,it->second);
				hashMapClassInfo->addObject(env, hashMap, key, vecObj);
			}

			return hashMap;
		}
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::FromGeoJSONAssembly()");
	}

	return NULL;
}

JNIEXPORT jobject JNICALL Java_com_mousebird_maply_VectorObject_getAttributes
  (JNIEnv *env, jobject obj)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		VectorObject *vecObj = classInfo->getObject(env,obj);
		if (!vecObj)
			return NULL;

		Dictionary *dict = vecObj->getAttributes();
		if (!dict)
		  return NULL;
		jobject dictObj = MakeAttrDictionary(env,dict);
	        return dictObj;
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::getAttributes()");
	}
    
    return NULL;
}

JNIEXPORT jobject JNICALL Java_com_mousebird_maply_VectorObject_centroid
  (JNIEnv *env, jobject obj)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		VectorObject *vecObj = classInfo->getObject(env,obj);
		if (!vecObj)
			return NULL;

		Point2f center;
		if (vecObj->centroid(center))
		{
			return MakePoint2d(env,Point2d(center.x(),center.y()));
		} else
			return NULL;
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::centroid()");
	}

    return NULL;
}

JNIEXPORT jobject JNICALL Java_com_mousebird_maply_VectorObject_largestLoopCenter
  (JNIEnv *env, jobject obj, jobject llObj, jobject urObj)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		Point2dClassInfo *pt2dClassInfo = Point2dClassInfo::getClassInfo();
		VectorObject *vecObj = classInfo->getObject(env,obj);
		Point2d *ll = pt2dClassInfo->getObject(env,llObj);
		Point2d *ur = pt2dClassInfo->getObject(env,urObj);
		if (!vecObj || !ll || !ur)
			return NULL;

		Point2f center,ll2f,ur2f;
		if (vecObj->largestLoopCenter(center,ll2f,ur2f))
		{
			*ll = Point2d(ll2f.x(),ll2f.y());
			*ur = Point2d(ur2f.x(),ur2f.y());
			return MakePoint2d(env,Point2d(center.x(),center.y()));
		} else
			return NULL;
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::largestLoopCenter()");
	}

    return NULL;
}

JNIEXPORT jdouble JNICALL Java_com_mousebird_maply_VectorObject_linearMiddle
  (JNIEnv *env, jobject obj, jobject midObj)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		Point2dClassInfo *pt2dClassInfo = Point2dClassInfo::getClassInfo();
		VectorObject *vecObj = classInfo->getObject(env,obj);
		Point2d *mid = pt2dClassInfo->getObject(env,midObj);
		if (!vecObj || !mid)
			return 0.0;

		Point2f mid2f;
		float rot;
		if (vecObj->linearMiddle(mid2f,rot))
		{
			*mid = Point2d(mid2f.x(),mid2f.y());
			return rot;
		} else
			return 0.0;
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::linearMiddle()");
	}
    
    return 0.0;
}

JNIEXPORT jboolean JNICALL Java_com_mousebird_maply_VectorObject_pointInside
(JNIEnv *env, jobject obj, jobject ptObj)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        Point2dClassInfo *pt2dClassInfo = Point2dClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        Point2d *pt = pt2dClassInfo->getObject(env,ptObj);
        if (!vecObj || !pt)
            return false;
        
        return vecObj->pointInside(*pt);

        return false;
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::pointInside()");
    }
    
    return false;
}

JNIEXPORT jint JNICALL Java_com_mousebird_maply_VectorObject_countPoints
(JNIEnv *env, jobject obj)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        Point2dClassInfo *pt2dClassInfo = Point2dClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        if (!vecObj)
            return 0;

        int numPts = 0;
        for (auto shape : vecObj->shapes)
        {
            if(std::dynamic_pointer_cast<VectorLinear>(shape) != NULL)
            {
                VectorLinearRef linear = std::dynamic_pointer_cast<VectorLinear>(shape);
                numPts += linear->pts.size();
            } else if(std::dynamic_pointer_cast<VectorLinear3d>(shape) != NULL)
            {
            } else if(std::dynamic_pointer_cast<VectorAreal>(shape) != NULL)
            {
                VectorArealRef ar = std::dynamic_pointer_cast<VectorAreal>(shape);
                if (ar)
                {
                    for (int ii=0;ii<ar->loops.size();ii++)
                    {
                        numPts += ar->loops[ii].size();
                    }
                }
            } else if(std::dynamic_pointer_cast<VectorPoints>(shape) != NULL)
            {
                VectorPointsRef points = std::dynamic_pointer_cast<VectorPoints>(shape);
                numPts += points->pts.size();
            }
        }
        
        return numPts;
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::countPoints()");
    }
    
    return false;
}

JNIEXPORT jboolean JNICALL Java_com_mousebird_maply_VectorObject_readFromFile
  (JNIEnv *env, jobject obj, jstring fileNameStr)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		VectorObject *vecObj = classInfo->getObject(env,obj);
		if (!vecObj)
			return false;
		const char *cStr = env->GetStringUTFChars(fileNameStr,0);
		if (!cStr)
			return false;
		std::string fileName(cStr);
		env->ReleaseStringUTFChars(fileNameStr, cStr);

		return vecObj->fromFile(fileName);
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::readFromFile()");
	}
    
    return false;
}

JNIEXPORT jboolean JNICALL Java_com_mousebird_maply_VectorObject_writeToFile
  (JNIEnv *env, jobject obj, jstring fileNameStr)
{
	try
	{
		VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
		VectorObject *vecObj = classInfo->getObject(env,obj);
		if (!vecObj)
			return false;
		const char *cStr = env->GetStringUTFChars(fileNameStr,0);
		if (!cStr)
			return false;
		std::string fileName(cStr);
		env->ReleaseStringUTFChars(fileNameStr, cStr);

		return vecObj->toFile(fileName);
	}
	catch (...)
	{
		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::writeToFile()");
	}
    
    return false;
}

JNIEXPORT jboolean JNICALL Java_com_mousebird_maply_VectorObject_tesselateNative
(JNIEnv *env, jobject obj, jobject retObj)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        VectorObject *retVecObj = classInfo->getObject(env,retObj);
        if (!vecObj || !retVecObj)
            return false;
        
        for (ShapeSet::iterator it = vecObj->shapes.begin();it!=vecObj->shapes.end();it++)
        {
            VectorArealRef ar = std::dynamic_pointer_cast<VectorAreal>(*it);
            if (ar)
            {
                VectorTrianglesRef trisRef = VectorTriangles::createTriangles();
                TesselateLoops(ar->loops, trisRef);
                trisRef->setAttrDict(*(ar->getAttrDict()));
                trisRef->initGeoMbr();
                retVecObj->shapes.insert(trisRef);
            }
        }
        
        return true;
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::tesselateNative()");
    }
    
    return false;
}

JNIEXPORT jboolean JNICALL Java_com_mousebird_maply_VectorObject_clipToGridNative
(JNIEnv *env, jobject obj, jobject retObj, jdouble sizeX, jdouble sizeY)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        VectorObject *retVecObj = classInfo->getObject(env,retObj);
        if (!vecObj || !retVecObj)
            return false;
        
        for (ShapeSet::iterator it = vecObj->shapes.begin();it!=vecObj->shapes.end();it++)
        {
            VectorArealRef ar = std::dynamic_pointer_cast<VectorAreal>(*it);
            if (ar)
            {
                std::vector<VectorRing> newLoops;
                ClipLoopsToGrid(ar->loops, Point2f(0.0,0.0), Point2f(sizeX,sizeY), newLoops);
                for (unsigned int jj=0;jj<newLoops.size();jj++)
                {
                    VectorArealRef newAr = VectorAreal::createAreal();
                    newAr->setAttrDict(*(ar->getAttrDict()));
                    newAr->loops.push_back(newLoops[jj]);
                    retVecObj->shapes.insert(newAr);
                }
            }
        }
        
        return true;
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::clipToGridNative()");
    }
    
    return false;
}

JNIEXPORT jboolean JNICALL Java_com_mousebird_maply_VectorObject_subdivideToGlobeNative
(JNIEnv *env, jobject obj, jobject retObj, jdouble epsilon)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        VectorObject *retVecObj = classInfo->getObject(env,retObj);
        if (!vecObj || !retVecObj)
            return false;
        
        retVecObj->shapes = vecObj->shapes;
        retVecObj->subdivideToGlobe(epsilon);
        
        return true;
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::subdivideToGlobeNative()");
    }
    
    return false;
}

JNIEXPORT jboolean JNICALL Java_com_mousebird_maply_VectorObject_subdivideToGlobeGreatCircleNative
(JNIEnv *env, jobject obj, jobject retObj, jdouble epsilon)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        VectorObject *retVecObj = classInfo->getObject(env,retObj);
        if (!vecObj || !retVecObj)
            return false;
        
        retVecObj->shapes = vecObj->shapes;
        retVecObj->subdivideToGlobeGreatCircle(epsilon);
        
        return true;
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::subdivideToGlobeGreatCircleNative()");
    }
    
    return false;
}

JNIEXPORT jboolean JNICALL Java_com_mousebird_maply_VectorObject_subdivideToFlatGreatCircleNative
(JNIEnv *env, jobject obj, jobject retObj, jdouble epsilon)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        VectorObject *retVecObj = classInfo->getObject(env,retObj);
        if (!vecObj || !retVecObj)
            return false;
        
        retVecObj->shapes = vecObj->shapes;
        retVecObj->subdivideToFlatGreatCircle(epsilon);
        
        return true;
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::subdivideToFlatGreatCircleNative()");
    }
    
    return false;
}


JNIEXPORT jboolean JNICALL Java_com_mousebird_maply_VectorObject_clipToMbrNative
(JNIEnv *env, jobject obj, jobject retObj, jdouble llX, jdouble llY, jdouble urX, jdouble urY)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        VectorObject *retVecObj = classInfo->getObject(env,retObj);
        if (!vecObj || !retVecObj)
            return false;

		Mbr mbr(Point2f(llX, llY), Point2f(urX, urY));

        for (ShapeSet::iterator it = vecObj->shapes.begin();it!=vecObj->shapes.end();it++)
        {
            if(std::dynamic_pointer_cast<VectorLinear>(*it) != NULL)
            {
	            VectorLinearRef linear = std::dynamic_pointer_cast<VectorLinear>(*it);
	            std::vector<VectorRing> newLoops;
	            ClipLoopToMbr(linear->pts, mbr, false, newLoops);
	            for (std::vector<VectorRing>::iterator it = newLoops.begin(); it != newLoops.end(); it++)
	            {
	                VectorLinearRef newLinear = VectorLinear::createLinear();
	                newLinear->setAttrDict(*(linear->getAttrDict()));
	                newLinear->pts = *it;
	                retVecObj->shapes.insert(newLinear);
	            }
            }
            else if(std::dynamic_pointer_cast<VectorLinear3d>(*it) != NULL)
            {
        		__android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Don't know how to clip linear3d objects");
				return false;
            }
            else if(std::dynamic_pointer_cast<VectorAreal>(*it) != NULL)
            {
	            VectorArealRef ar = std::dynamic_pointer_cast<VectorAreal>(*it);
	            if (ar)
	            {
	                for (int ii=0;ii<ar->loops.size();ii++)
	                {
	                    std::vector<VectorRing> newLoops;
	                    ClipLoopToMbr(ar->loops[ii], mbr, true, newLoops);
	                    for (unsigned int jj=0;jj<newLoops.size();jj++)
	                    {
	                        VectorArealRef newAr = VectorAreal::createAreal();
	                        newAr->setAttrDict(*(ar->getAttrDict()));
	                        newAr->loops.push_back(newLoops[jj]);
	                        retVecObj->shapes.insert(newAr);
	                    }
	                }
	            }
            }
            else if(std::dynamic_pointer_cast<VectorPoints>(*it) != NULL)
            {
	            VectorPointsRef points = std::dynamic_pointer_cast<VectorPoints>(*it);
	            VectorPointsRef newPoints = VectorPoints::createPoints();
	            for (unsigned int ii=0;ii<points->pts.size();ii++)
	            {
	                const Point2f &pt = points->pts[ii];
	                if(pt.x() >= llX && pt.x() <= urX &&
	                   pt.y() >= llY && pt.y() <= urY)
	                {
	                    newPoints->pts.push_back(pt);
	                }
	            }
            }
        }

        return true;
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::clipToMbrNative()");
    }
    
    return false;
}

JNIEXPORT jint JNICALL Java_com_mousebird_maply_VectorObject_getVectorTypeNative
(JNIEnv *env, jobject obj)
{
    try
    {
        VectorObjectClassInfo *classInfo = VectorObjectClassInfo::getClassInfo();
        VectorObject *vecObj = classInfo->getObject(env,obj);
        if (!vecObj)
            return MaplyVectorNoneType;
		return vecObj->getVectorType();
    }
    catch (...)
    {
        __android_log_print(ANDROID_LOG_VERBOSE, "Maply", "Crash in VectorObject::clipToMbrNative()");
    }
	return MaplyVectorNoneType;
}
